/**
 * gtk_helloworld-2.c
 *
 * This maemo code example is licensed under a MIT-style license,
 * that can be found in the file called "License" in the same
 * directory as this file.
 * Copyright (c) 2007 Nokia Corporation. All rights reserved.
 *
 * This version adds proper mechanisms to end the program.
 *
 * Look for lines with "NEW" or "MODIFIED" in them.
 */

#include <stdlib.h>
#include <gtk/gtk.h>

/**
 * NEW
 *
 * Callback function (event handler) for the "delete" event.
 * This event is emitted by GTK+ main loop after a window manager has
 * requested this window to be closed.
 *
 * NOTES:
 * - This is really a signal handler.
 * - The first parameter to a signal handler is a pointer to the
 *   object that caused the signal to be emitted (when using
 *   gobject_signal_connect).
 * - There are zero or more parameters in between the first one and
 *   the last one. The count of parameters depends on the signal
 *   type.
 * - The last parameter is always an untyped pointer (again, when
 *   using gobject_signal_connect).
 *
 * Parameters:
 * - widget: Pointer to the widget that emitted the delete-event.
 * - event:  GdkEvent-structure describing what kind of event this
 *           is. We'll register this handler to handle only the
 *           "destroy-event", so we'll know what kind of event this
 *           is without checking the "event"-parameter.
 * - data:   Untyped pointer for passing data between the main loop
 *           and handlers. We ignore it for now.
 *
 * Returns:
 * - gboolean: Whether the event has been consumed by this handler or
 *             not.
 * NOTE:
 *   Signal handlers may exit with an return value. Whether this
 *   value will be used by anyone, depends on the signal type. For
 *   events, we must always return a boolean value.
 */
static gboolean delete_event(GtkWidget* widget, GdkEvent event,
                                                gpointer data) {
  /* Print out an informational message. */
  g_print("CB:delete_event\n");

  /* We want to tell the gtk_main that we're done with this widget.
     Since it's the top-level widget in our window (it is the
     window), returning FALSE to our caller signifies that the event
     has not been processed by us, and this in turn means that it
     will be propagated to the default window delete-event handler
     which will terminate the program. This is done in by the caller
     so that it will issue a "destroy"-signal next. If we would
     return TRUE, the program wouldn't stop. This behavior is
     specific to the "delete-event" signal. */
  return FALSE;
}

/**
 * NEW
 *
 * Our callback function for the "destroy"-signal which is issued
 * when the Widget is going to be destroyed.
 * NOTE:
 *   In this case we don't need to return anything, since we're just
 *   processing a simple signal.
 *
 * This is used by widgets that are going away that want to notify us
 * that they will not be around for much longer. They will remove
 * themselves eventually.
 *
 * We call gtk_main_quit() which is a routine that will tell gtk_main
 * that it's time to terminate.
 *
 * Note, you must not terminate your program here! (do not use the
 * exit()-library function). This would leave the GTK+ library in a
 * slightly confused state and might leave other threads hanging
 * around. In this material we won't be using threads, but consider
 * yourself warned.
 *
 * Parameters:
 * - widget: Same as with event handler callbacks.
 * - data:   Same as with event handler callbacks.
 *
 * Returns:
 * void
 */
static void end_program(GtkWidget* widget, gpointer data) {
  /* Print out an informational message to stdout. */
  g_print("CB:end_program: calling gtk_main_quit()\n");

  /* Tell gtk_main to terminate the application. */
  gtk_main_quit();

  /* Display message about what is happening. */
  g_print("CB:end_program: back from gtk_main_quit & returning\n");
}

/**
 * MODIFIED
 *
 * We connect the signals to the handlers so that the application may
 * be terminated properly from the GUI.
 */
int main(int argc, char** argv) {

  /* We'll have two references to two GTK+ widgets. */
  GtkWindow* window;
  GtkLabel* label;

  /* Initialize the GTK+ library. */
  gtk_init(&argc, &argv);

  /* Create a window with window border width of 12 pixels and a
     title text. */
  window = g_object_new(GTK_TYPE_WINDOW,
    "border-width", 12,
    "title", "Hello GTK+",
    NULL);

  /* Create the label widget. */
  label = g_object_new(GTK_TYPE_LABEL,
    "label", "Hello World!",
    NULL);

  /* Pack the label into the window layout. */
  gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(label));

  /**
   * NEW
   *
   * Here we "install" event and signal handlers.
   *
   * Events are handled by the GTK+ as signals. Their difference is
   * explained in the text that follows.
   *
   * g_signal_connect(GtkWidget*, gchar*, CB*, gpointer*) is the API
   * function to connect signals from some object to some signal
   * receiver (callback function when using C).
   *
   * The second parameter should point to a string giving the event
   * or signal name. GLib treats underscores and dashes in signal
   * names as they would be the same character. The API documentation
   * uses the dash (-) and so will we. Needless to say, different
   * kind of widgets are capable of emitting different events.
   *
   * Third parameter is the address of the callback function that
   * should receive the signal. We use G_CALLBACK()-macros to force a
   * typecast into the type specification that g_signal_connect
   * expects. Depending on the signal the callback will receive
   * different number of parameters and is expected to return
   * something in some cases. There is no hard and set rule.
   *
   * THIS MEANS THAT YOU HAVE TO BE SURE ABOUT THE CAST VALIDITY.
   *
   * It is possible to pass an extra pointer (the last parameter)
   * that will be carried with the signal to the signal receiver. We
   * will see uses for this later, but in real applications it is
   * quite important. This last parameter is defined as a gpointer
   * (untyped pointer, void*), and if one needs to pass other data
   * in the pointer's place, one will need to do additional
   * typecasting (we'll see this later).
   */
  g_signal_connect(window, "delete-event",
                   G_CALLBACK(delete_event), NULL);
  g_signal_connect(window, "destroy",
                   G_CALLBACK(end_program), NULL);

  /* Show all widgets that are contained by the window. */
  gtk_widget_show_all(GTK_WIDGET(window));

  /* Start the main event loop. */
  g_print("main: calling gtk_main\n");
  gtk_main();

  /* Display a message to the standard output and exit. */
  g_print("main: returned from gtk_main and exiting with success\n");

  /* Return success as exit code. */
  return EXIT_SUCCESS;
}
