/**
 * gtk_helloworld-3.c
 *
 * This maemo code example is licensed under a MIT-style license,
 * that can be found in the file called "License" in the same
 * directory as this file.
 * Copyright (c) 2007 Nokia Corporation. All rights reserved.
 *
 * We add a menu and callbacks to process menu selections.
 *
 * Look for lines with "NEW" or "MODIFIED" in them.
 */

#include <stdlib.h>
#include <gtk/gtk.h>

/**
 * NEW
 *
 * We will use a single callback function to process the activation
 * of all of the menu items.
 *
 * We could also use strings to identify which menu item was
 * selected, or if our callback would have a list of pointers to the
 * menu items themselves, we could compare against that. Instead,
 * we'll define an enumeration that will contain all the possible
 * menu items that our application will have. This allows use to use
 * a simple switch statement to determine what to do.
 */
typedef enum {
  MENU_FILE_OPEN = 1,
  MENU_FILE_SAVE = 2,
  MENU_FILE_QUIT = 3
} MenuActionCode;

/**
 * Callback function (event handler) for the "delete" event.
 */
static gboolean delete_event(GtkWidget* widget, GdkEvent event,
                                                gpointer data) {
  /* Print out an informational message. */
  g_print("CB:delete_event\n");

  /* Done with the widget. */
  return FALSE;
}

/**
 * Our callback function for the "destroy"-signal which is issued
 * when the Widget is going to be destroyed.
 */
static void end_program(GtkWidget* widget, gpointer data) {
  /* Print out an informational message to stdout. */
  g_print("CB:end_program: calling gtk_main_quit()\n");

  /* Tell gtk_main to terminate the application. */
  gtk_main_quit();

  /* Display message about what is happening. */
  g_print("CB:end_program: back from gtk_main_quit & returning\n");
}

/**
 * NEW
 *
 * Signal handler for the menu item selections.
 *
 * Gets the menu action code as a parameter but since we'll cast it
 * into an pointer on callback registration, we'll need to cast it
 * back internally.
 */
static void cbActivation(GtkMenuItem* mi, gpointer data) {

  /* Convert the pointer into an integer, irrespective of the pointer
     size. */
  MenuActionCode aCode = GPOINTER_TO_INT(data);

  switch(aCode) {
    case MENU_FILE_OPEN:
      g_print("Selected open\n");
      break;
    case MENU_FILE_SAVE:
      g_print("Selected save\n");
      break;
    case MENU_FILE_QUIT:
      g_print("Selected quit\n");
      gtk_main_quit();
      break;
    default:
      /* Besides g_print, there are two other output functions in
         GLib: g_warning and g_error. g_error will stop our program
         (using abort() internally), so we'll choose to use g_warning
         instead since we want our program to continue running. The
         situation is not lethal, but should be investigated.

         g_warning prints out a warning message in the same format as
         g_print and printf. */
      g_warning("cbActivation: Unknown menu action code %d.\n", aCode);
  }
}

/**
 * NEW
 *
 * A convenience function to create menu items with internal label
 * objects.
 *
 * NOTE:
 *   Most existing programs use a convenience function that GTK+
 *   provides (gtk_menu_item_new_with_label).
 *
 * We want to use the GObject API for this example to demonstrate its
 * compactness.
 *
 * Parameters:
 * - const gchar*: string to display in the label inside the item.
 *
 * Returns:
 * - A pointer to a new MenuItem that holds the label
 */
static GtkMenuItem* buildMenuItem(const gchar* labelText) {

  GtkLabel* label;
  GtkMenuItem* mi;

  /* Create the Label object first. We set the text using its "label"
     property. We also set its alignment using a property that its
     superclass (GtkMisc) has. Namely "xalign". xalign with 0.0 is
     left and 1.0 is right extreme. Default value for xalign is 0.5
     which would align the text using center justification. */
  label = g_object_new(GTK_TYPE_LABEL,
    "label", labelText,     /* GtkLabel property */
    "xalign", (gfloat)0.0,  /* GtkMisc property */
    NULL);

  /* Next we create the GtkMenuItem.

     A MenuItem is a container that can hold one child (this
     restriction comes from GtkBin-class, which is an abstract class
     for containers that are capable of holding only one one child).
     We use the "child" property of the Container class to set its
     child on creation. Normally a Container would allow us to add
     multiple children in a similar way (by setting the "child"-
     property multiple times). */
  mi = g_object_new(GTK_TYPE_MENU_ITEM, "child", label, NULL);

  /* Return the GtkMenuItem to caller. */
  return mi;
}

/**
 * NEW
 *
 * Utility function to build a menubar and the menu.
 *
 * Uses the buildMenuItem-function from above and will connect the
 * signals that can be generated by MenuItems when they're selected
 * (the "activate"-signal).
 *
 * Returns:
 * - the completed MenuBar
 */
static GtkMenuBar* buildMenubar(void) {

  /* The visible menubar widget that will hold one submenu called
     "File". */
  GtkMenuBar* menubar;
  /* The menu item representing the menu-name in the menubar. It is
     necessary in order to attach the menu to the menubar. Will open
     the submenu when activated. */
  GtkMenuItem* miFile;

  /* A container implementing a menu that holds multiple items. */
  GtkMenu* fileMenu;
  /* The individual menu items and a separator item. */
  GtkMenuItem* miOpen;
  GtkMenuItem* miSave;
  GtkMenuItem* miSep;
  GtkMenuItem* miQuit;

  /* Create the container for items. */
  fileMenu = g_object_new(GTK_TYPE_MENU, NULL);

  /* Create the menu items. */
  miOpen = buildMenuItem("Open");
  miSave = buildMenuItem("Save");
  miQuit = buildMenuItem("Quit");
  miSep = g_object_new(GTK_TYPE_SEPARATOR_MENU_ITEM, NULL);

  /* Add the items into the container.

     Here we again use the "child"-property. But since Menu is a
     proper container that can hold multiple children, we can use a
     list of widgets to add. */
  g_object_set(fileMenu,
    "child", miOpen,
    "child", miSave,
    "child", miSep,
    "child", miQuit,
    NULL);

  /* Connect the signals from the individual menu items.
     NOTE:
       We need to cast the CB data into a gpointer since that is what
       the prototype of g_signal_connect tells the compiler to
       enforce. Luckily for us there is a macro to do exactly this
       (in GLib). */
  g_signal_connect(G_OBJECT(miOpen), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_OPEN));
  g_signal_connect(G_OBJECT(miSave), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_SAVE));
  g_signal_connect(G_OBJECT(miQuit), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_QUIT));

  /* Create the menu item that will be added to the menubar. */
  miFile = buildMenuItem("File");

  /* Associate the "File" menu item to the Menu container so that it
     will open it as a submenu when activated.

     There is no property to set this, so we use an accessor
     function. */
  gtk_menu_item_set_submenu(miFile, GTK_WIDGET(fileMenu));

  /* Create the menubar. */
  menubar = g_object_new(GTK_TYPE_MENU_BAR, NULL);

  /* Add the "File" menu item to the menubar.

     We want to add from left to right, so again, the "child"
     property will be used. */
  g_object_set(menubar, "child", miFile, NULL);

  /* Return the completed menubar to the caller. */
  return menubar;
}

/**
 * MODIFIED
 *
 * We add a container that can hold more than one widget since
 * GtkWindow can hold only one at a time. We need to put our menubar
 * somewhere, and display a label. We select a container that will
 * split the area that it will get into horizontal sections according
 * to the needs to its children (GtkVBox).
 */
int main(int argc, char** argv) {

  GtkWindow* window;
  GtkLabel* label;
  GtkMenuBar* menubar;
  GtkVBox* vbox;

  /* Initialize the GTK+ library. */
  gtk_init(&argc, &argv);

  /* Create a window with window border width of 12 pixels and a
     title text. */
  window = g_object_new(GTK_TYPE_WINDOW,
    "border-width", 12,
    "title", "Hello GTK+",
    NULL);

  /* Create the label widget. */
  label = g_object_new(GTK_TYPE_LABEL,
    "label", "Hello World! (with menus)",
    NULL);

  /* Build the menubar (NEW). */
  menubar = buildMenubar();

  /* Create a layout box for the window since it can only hold one
     widget (NEW). */
  vbox = g_object_new(GTK_TYPE_VBOX, NULL);

  /* Add the vbox as the only child for the window (NEW). */
  g_object_set(window, "child", vbox, NULL);

  /* NEW

     Add it to the container that now controls the area inside the
     window.

     Both VBox and HBox implement an abstract class called GtkBox.
     Adding child widgets to GtkBoxes is possible by using an
     old-style API (gtk_box_pack_start and _end), or we can use the
     context based one with properties.

     We need to control three properties (besides the "child", GtkBox
     is also a container):
     - "pack-type": selects whether to add the next child to the
       start or end of box. Default is GTK_PACK_START.
       For VBox start is top and end is bottom.
     - "expand": Controls whether the child will receive extra space
       when the Box will grow (if it get's new space itself). Default
       is TRUE.
     - "fill": If "expand" is set, controls whether extra new space
       should be given to the child or used as padding. Default is
       TRUE.

    Based on this information we now must select a sequence that will
    create a split between the menubar and the label.

    We want the menubar to take only the amount of space that it
    requires, and for the label to fill up the rest. Should the
    window grow, we want the area allocated to label to grow as well.
    We'll see a growing window later on.

    We have two choices (A or B):
    A) Insert the menubar first:
       1) Set "expand" to FALSE.
       2) Use "child" to insert menubar.
       3) Set "pack-type" to GTK_PACK_END so that label will be added
          from bottom-up.
       4) Set "expand" back to TRUE.
       5) Use "child" to insert the label.
    B) Insert the Label first:
       1) Set "pack-type" to GTK_PACK_END.
       2) Insert label using "child".
       3) Set "expand" to FALSE.
       4) Set "pack-type" to GTK_PACK_START.
       5) Use "child" to insert the menubar.

    As you can see, both approaches will require an equal amount of
    property operations, so we can choose either one. There are other
    possible orders as well, but all of them require at least the
    five operations. We'll choose the first approach (A):

    g_object_set(vbox,
      "fill", FALSE,
      "child", menubar,
      "pack-type", GTK_PACK_END,
      "expand", TRUE,
      "child", label,
      NULL);

    NOTE:
      This won't work currently since the VBox implementation does
      not implement the properties of the abstract GtkBox class.

      This means that we'll need to use the accessor interface which
      requires us to give both fill and expand each time we add a
      widget.

      The order of parameters for the accessor function:
      - GtkBox into which we're adding the widget.
      - The widget that we're adding.
      - "fill".
      - "expand".
      - "padding", which is how many pixels are used around the
        widget. */
  gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(menubar), FALSE,
                     FALSE, 0);
  gtk_box_pack_end(GTK_BOX(vbox), GTK_WIDGET(label), TRUE, TRUE, 0);

  /* Connect the termination signals */
  g_signal_connect(window, "delete-event",
                   G_CALLBACK(delete_event), NULL);
  g_signal_connect(window, "destroy",
                   G_CALLBACK(end_program), NULL);

  /* Show all widgets that are contained by the window. This will
     also include the menubar, and its widget tree. */
  gtk_widget_show_all(GTK_WIDGET(window));

  /* Start the main event loop. */
  g_print("main: calling gtk_main\n");
  gtk_main();

  /* Display a message to the standard output and exit. */
  g_print("main: returned from gtk_main and exiting with success\n");

  /* Return success as exit code. */
  return EXIT_SUCCESS;
}

