/**
 * hildon_helloworld-1.c
 *
 * This maemo code example is licensed under a MIT-style license,
 * that can be found in the file called "License" in the same
 * directory as this file.
 * Copyright (c) 2007 Nokia Corporation. All rights reserved.
 *
 * Let's hildonize our application. Most of the application stays
 * unchanged, but the menu building utility needs some changes and
 * 'main' will need to be modified.
 *
 * Look for lines with "NEW" or "MODIFIED" in them.
 */

#include <stdlib.h>
#include <gtk/gtk.h>

/**
 * NEW
 *
 * When using widgets from the Hildon library, we need to use
 * additional include files. We could just pull in <hildon/hildon.h>,
 * but that is unnecessary since we only use a small number of Hildon
 * widgets. Instead we'll opt to pull in only the Hildon header files
 * that are necessary for the widgets that we use.
 *
 * For our first Hildon application, we'll need the Hildon Program
 * framework widgets whose API and types are declared in
 * hildon-program.h
 */
#include <hildon/hildon-program.h>

/* Menu codes. */
typedef enum {
  MENU_FILE_OPEN = 1,
  MENU_FILE_SAVE = 2,
  MENU_FILE_QUIT = 3
} MenuActionCode;

/**
 * Callback function (event handler) for the "delete" event.
 */
static gboolean delete_event(GtkWidget* widget, GdkEvent event,
                                                gpointer data) {
  return FALSE;
}

/**
 * Our callback function for the "destroy"-signal which is issued
 * when the Widget is going to be destroyed.
 */
static void end_program(GtkWidget* widget, gpointer data) {
  gtk_main_quit();
}

/**
 * Signal handler for the menu item selections.
 */
static void cbActivation(GtkMenuItem* mi, gpointer data) {

  MenuActionCode aCode = GPOINTER_TO_INT(data);

  switch(aCode) {
    case MENU_FILE_OPEN:
      g_print("Selected open\n");
      break;
    case MENU_FILE_SAVE:
      g_print("Selected save\n");
      break;
    case MENU_FILE_QUIT:
      g_print("Selected quit\n");
      gtk_main_quit();
      break;
    default:
      g_warning("cbActivation: Unknown menu action code %d.\n", aCode);
  }
}

/**
 * A convenience function to create menu items with internal label
 * objects.
 */
static GtkMenuItem* buildMenuItem(const gchar* labelText) {

  GtkLabel* label;
  GtkMenuItem* mi;

  /* Create the Label object. */
  label = g_object_new(GTK_TYPE_LABEL,
    "label", labelText,     /* GtkLabel property */
    "xalign", (gfloat)0.0,  /* GtkMisc property */
    NULL);

  /* Create the GtkMenuItem and add the Label as its child. */
  mi = g_object_new(GTK_TYPE_MENU_ITEM, "child", label, NULL);

  /* Return the GtkMenuItem to caller. */
  return mi;
}

/**
 * MODIFIED
 *
 * This utility adds the menu items into the given HildonProgram.
 * It uses the buildMenuItem-function internally to build each menu
 * item. It will connect the signals that can be generated by
 * MenuItems when they're selected ("activate"-signal).
 *
 * Since there is no menubar, so this becomes quite simple.
 */
static void buildMenu(HildonProgram* program) {

  GtkMenu* menu;
  GtkMenuItem* miOpen;
  GtkMenuItem* miSave;
  GtkMenuItem* miSep;
  GtkMenuItem* miQuit;

  /* Create the menu items. */
  miOpen = buildMenuItem("Open");
  miSave = buildMenuItem("Save");
  miQuit = buildMenuItem("Quit");
  miSep = g_object_new(GTK_TYPE_SEPARATOR_MENU_ITEM, NULL);

  /* Create a new menu. */
  menu = g_object_new(GTK_TYPE_MENU, NULL);

  /* Add the items to the container. */
  g_object_set(menu,
    "child", miOpen,
    "child", miSave,
    "child", miSep,
    "child", miQuit,
    NULL);

  /* Set the top-level menu for Hildon program (NEW). */
  hildon_program_set_common_menu(program, menu);

  /* Connect the signals from the individual menu items. */
  g_signal_connect(G_OBJECT(miOpen), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_OPEN));
  g_signal_connect(G_OBJECT(miSave), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_SAVE));
  g_signal_connect(G_OBJECT(miQuit), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_QUIT));

  /**
   * NEW
   *
   * We need to explicitly set the visibility for the menu since by
   * default it's not visible when we get it from window. Safest bet
   * is to show the widget tree that is inside the menu.
   */
  gtk_widget_show_all(GTK_WIDGET(menu));
}

/**
 * MODIFIED
 *
 * We switch to the application framework that Hildon provides.
 */
int main(int argc, char** argv) {

  /* Our GtkWindow has been replaced with HildonProgram (NEW). */
  HildonProgram* program;
  /* And we'll have one window, which will act as the container for
     our needs (NEW). */
  HildonWindow* window;
  GtkWidget* label;
  GtkWidget* vbox;

  /* Initialize the GTK+. */
  gtk_init(&argc, &argv);

  /**
   * NEW
   *
   * Create the Hildon provided window and controller widgets.
   */
  program = HILDON_PROGRAM(hildon_program_get_instance());
  /* Set the application title using an accessor function. */
  g_set_application_name("Hello Hildon!");
  /* Create a window that will handle our layout and menu. */
  window = HILDON_WINDOW(hildon_window_new());
  /* Bind the HildonWindow to HildonProgram. */
  hildon_program_add_window(program, HILDON_WINDOW(window));

  /* Create the label widget. */
  label = g_object_new(GTK_TYPE_LABEL,
    "label", "Hello Hildon (with menus)!",
    NULL);

  /* Build the menu. We'll create the menu for the HildonProgram, so
     we pass the pointer to our revised buildMenu (NEW). */
  buildMenu(program);

  /* Create a layout box for the window since it can only hold one
     widget. */
  vbox = g_object_new(GTK_TYPE_VBOX, NULL);

  /* Add the vbox as a child to the Window. */
  g_object_set(window, "child", vbox, NULL);

  /* Pack the label into the VBox. */
  gtk_box_pack_end(GTK_BOX(vbox), GTK_WIDGET(label), TRUE, TRUE, 0);

  /* Connect the termination signals. Note how the HildonWindow
     object has taken the responsibilities that a GtkWindow normally
     would have (NEW). */
  g_signal_connect(G_OBJECT(window), "delete-event",
                   G_CALLBACK(delete_event), NULL);
  g_signal_connect(G_OBJECT(window), "destroy",
                   G_CALLBACK(end_program), NULL);

  /* Show all widgets that are contained by the window. This also
     includes the menu (if it has been setup by this point) (NEW). */
  gtk_widget_show_all(GTK_WIDGET(window));

  /* Start the main event loop. */
  g_print("main: calling gtk_main\n");
  gtk_main();

  g_print("main: returned from gtk_main and exiting with success\n");

  return EXIT_SUCCESS;
}
