/**
 * hildon_helloworld-3.c
 *
 * This maemo code example is licensed under a MIT-style license,
 * that can be found in the file called "License" in the same
 * directory as this file.
 * Copyright (c) 2007 Nokia Corporation. All rights reserved.
 *
 * We now add a submenu with radio and check menu items.
 *
 * Look for lines with "NEW" or "MODIFIED" in them.
 */

#include <stdlib.h>
#include <gtk/gtk.h>
#include <hildon/hildon-program.h>

/* Menu codes. */
typedef enum {
  MENU_FILE_OPEN = 1,
  MENU_FILE_SAVE = 2,
  MENU_FILE_QUIT = 3,
  /* NEW:

     We will have three new signals handled by our signal handler. */
  MENU_STYLE_UNDERLINE_TOGGLED = 4,
  MENU_STYLE_NORMAL_TOGGLED = 5,
  MENU_STYLE_ITALIC_TOGGLED = 6
} MenuActionCode;

/**
 * Callback function (event handler) for the "delete" event.
 */
static gboolean delete_event(GtkWidget* widget, GdkEvent event,
                                                gpointer data) {
  return FALSE;
}

/**
 * Our callback function for the "destroy"-signal which is issued
 * when the Widget is going to be destroyed.
 */
static void end_program(GtkWidget* widget, gpointer data) {
  gtk_main_quit();
}

/**
 * MODIFIED
 *
 * Signal handler for the menu item selections.
 */
static void cbActivation(GtkMenuItem* mi, gpointer data) {

  MenuActionCode aCode = GPOINTER_TO_INT(data);

  switch(aCode) {
    case MENU_FILE_OPEN:
      g_print("Selected open\n");
      break;
    case MENU_FILE_SAVE:
      g_print("Selected save\n");
      break;
    case MENU_FILE_QUIT:
      g_print("Selected quit\n");
      gtk_main_quit();
      break;
    case MENU_STYLE_UNDERLINE_TOGGLED:
      /* NEW

         Check items can be in two states. In order to get the
         current (new) state, we use an API function. The API
         function expects a GtkCheckMenuItem widget as its parameter,
         hence the typecast. */
      g_print("Style underline has been toggled. New state is: %s\n",
        gtk_check_menu_item_get_active(
          GTK_CHECK_MENU_ITEM(mi))?"on":"off");
      break;
    case MENU_STYLE_NORMAL_TOGGLED:
    case MENU_STYLE_ITALIC_TOGGLED:
      /* NEW

         We handle toggles for both radio items in the same code.
         This code will be called twice on each toggle since one of
         the radio items will be toggled off, the other will be
         toggled on.

         The order is as follows:
         1) Both go to 'off'-state
         2) One of the toggles goes into 'on'-state.

         Our logic is constructed as follows:
         1) Find out whether the radio item that emitted this signal
            is 'on' (active).
         2) If so, determine which of the radio items it is. For this
            we use another switch and then print the respective
            message.
         3) If not, we don't do anything since this signal will be
            shortly repeated from the other radio item. */
      {
        gboolean isActive = FALSE;

        isActive = gtk_check_menu_item_get_active(
          GTK_CHECK_MENU_ITEM(mi));
        /* This is only for debugging. aCode for NORMAL is 5, and for
           ITALIC it is 6. This will demonstrate that at some point,
           both radio items are "off" at the same time. */
        g_print(" style: aCode=%d isActive=%d\n", aCode, isActive);
        if (isActive) {
          g_print("Style 'slant' has changed: ");
          switch (aCode) {
            case MENU_STYLE_NORMAL_TOGGLED:
              g_print("Normal selected\n");
              break;
            case MENU_STYLE_ITALIC_TOGGLED:
              g_print("Italic selected\n");
              break;
            /* This keeps gcc from complaining. */
            default: break;
          }
        }
        break;
      }
    default:
      g_warning("cbActivation: Unknown menu action code %d.\n", aCode);
  }
}

/**
 * NEW
 *
 * We create a submenu which will contain three items:
 * - A checkbox item for selecting whether underlining is requested.
 * - A group of two radio items out of which only one can be selected
 *   at any given time (as is common for radio button groups).
 *
 * Returns:
 * - The sub-menu (as GtkWidget*)
 */
static GtkWidget* buildSubMenu(void) {

  GtkWidget* subMenu;

  GtkWidget* miUnderline;
  GtkWidget* miSep;
  GtkWidget* miItalic;
  GtkWidget* miNormal;

  /* Create a checkbox menu item. */
  miUnderline = gtk_check_menu_item_new_with_label("Underline");
  /* Set it as "checked" (default is FALSE). */
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(miUnderline),
    TRUE);

  /* Create the radio items.

     If you're wondering what this extra brace is doing here, don't
     worry :-). It's a scope so that we can define a variable and
     forget about it quickly when we leave the scope. You should
     check your style guidelines whether this is a good idea or not.

     It was used here so that we won't access group by mistake since
     its memory and reference counts are controlled by GTK+. */
  {
    GSList* group = NULL;

    /* Normally the first parameter would be the group into which
       this new radio menu item would be added. If we leave it as
       NULL, GTK+ will create a new group for that which is of the
       type GSList (GLib single-linked list). */
    miItalic = gtk_radio_menu_item_new_with_label(NULL, "Italic");
    /* Get the group so that we can add another radio menu item. */
    group = gtk_radio_menu_item_get_group(
            GTK_RADIO_MENU_ITEM(miItalic));
    /* Create another radio menu item and add it as well. */
    miNormal = gtk_radio_menu_item_new_with_label(group, "Normal");
    /* You will also sometimes see code like this:
      miItalic = gtk_rad..el(NULL, "Normal");
      miNormal = gtk_rad..el_from_widget(
        GTK_RADIO_MENU_ITEM(miItalic, "Normal"));

    This is a shortcut so that we don't need to get the group for the
    widget every time. For multiple radio menu items (or buttons
    too), this will get rather tedious. */
  }

  /* Create the separator item. */
  miSep = gtk_separator_menu_item_new();

  /* Create the menu to hold the items. */
  subMenu = gtk_menu_new();

  /* Add the items to the menu.

     If you see code like gtk_menu_append, re-write that to use this
     version since gtk_menu_append is deprecated. GtkMenuShell is an
     abstract superclass for menus-style widgets that can hold
     selectable child-widgets.

     You could also use gtk_container_add for adding the menu items
     but gtk_menu_shell also contains possibilities for using
     different orders for the items so it's useful to know that it
     exists. */
  gtk_menu_shell_append(GTK_MENU_SHELL(subMenu), miUnderline);
  gtk_menu_shell_append(GTK_MENU_SHELL(subMenu), miSep);
  gtk_menu_shell_append(GTK_MENU_SHELL(subMenu), miItalic);
  gtk_menu_shell_append(GTK_MENU_SHELL(subMenu), miNormal);

  /* Connect the signals. */
  g_signal_connect(G_OBJECT(miUnderline), "toggled",
    G_CALLBACK(cbActivation),
    GINT_TO_POINTER(MENU_STYLE_UNDERLINE_TOGGLED));
  g_signal_connect(G_OBJECT(miItalic), "toggled",
    G_CALLBACK(cbActivation),
    GINT_TO_POINTER(MENU_STYLE_ITALIC_TOGGLED));
  g_signal_connect(G_OBJECT(miNormal), "toggled",
    G_CALLBACK(cbActivation),
    GINT_TO_POINTER(MENU_STYLE_NORMAL_TOGGLED));

  /* The submenu is ready, return it to caller. */
  return subMenu;
}

/**
 * MODIFIED
 *
 * This utility creates the menu for the HildonProgram.
 */
static void buildMenu(HildonProgram* program) {

  GtkMenu* menu;
  GtkWidget* miOpen;
  GtkWidget* miSave;
  /* Menu item whichs opens the style submenu (NEW). */
  GtkWidget* miStyle;
  GtkWidget* subMenu;
  GtkWidget* miSep;
  GtkWidget* miQuit;

  /* Create the menu items. */
  miOpen = gtk_menu_item_new_with_label("Open");
  miSave = gtk_menu_item_new_with_label("Save");
  miStyle = gtk_menu_item_new_with_label("Style");
  miQuit = gtk_menu_item_new_with_label("Quit");
  miSep =  gtk_separator_menu_item_new();

  /* Build the submenu (NEW). */
  subMenu = buildSubMenu();

  /* Connect the style-item so that it will pop up the submenu (NEW). */
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(miStyle), subMenu);

  /* Create a new menu. */
  menu = GTK_MENU(gtk_menu_new());

  /* Add the items to the container. */
  gtk_container_add(GTK_CONTAINER(menu), miOpen);
  gtk_container_add(GTK_CONTAINER(menu), miSave);
  gtk_container_add(GTK_CONTAINER(menu), miStyle);
  gtk_container_add(GTK_CONTAINER(menu), miSep);
  gtk_container_add(GTK_CONTAINER(menu), miQuit);

  /* Connect the signals from individual menu items. */
  g_signal_connect(G_OBJECT(miOpen), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_OPEN));
  g_signal_connect(G_OBJECT(miSave), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_SAVE));
  g_signal_connect(G_OBJECT(miQuit), "activate",
    G_CALLBACK(cbActivation), GINT_TO_POINTER(MENU_FILE_QUIT));

  /* Set the top level menu for Hildon program. */
  hildon_program_set_common_menu(program, menu);
}

/**
 * The main program remains the same as before.
 */
int main(int argc, char** argv) {

  HildonProgram* program;
  HildonWindow* window;
  GtkWidget* label;
  GtkWidget* vbox;

  /* Initialize the GTK+. */
  gtk_init(&argc, &argv);

  /* Create the Hildon program. */
  program = HILDON_PROGRAM(hildon_program_get_instance());
  /* Set the application title using an accessor function. */
  g_set_application_name("Hello Hildon!");
  /* Create a window that will handle our layout and menu. */
  window = HILDON_WINDOW(hildon_window_new());
  /* Bind the HildonWindow to HildonProgram. */
  hildon_program_add_window(program, HILDON_WINDOW(window));

  /* Create the label widget. */
  label = gtk_label_new("Hello Hildon (with submenus)!");

  /* Build the menu and attach it to the HildonProgram. */
  buildMenu(program);

  /* Create a layout box for the window. */
  vbox = gtk_vbox_new(FALSE, 0);

  /* Add the vbox as a child to the Window. */
  gtk_container_add(GTK_CONTAINER(window), vbox);

  /* Pack the label into the VBox. */
  gtk_box_pack_end(GTK_BOX(vbox), GTK_WIDGET(label), TRUE, TRUE, 0);

  /* Connect the termination signals. */
  g_signal_connect(G_OBJECT(window), "delete-event",
                   G_CALLBACK(delete_event), NULL);
  g_signal_connect(G_OBJECT(window), "destroy",
                   G_CALLBACK(end_program), NULL);

  /* Show all widgets that are contained by the Window. */
  gtk_widget_show_all(GTK_WIDGET(window));

  /* Start the main event loop. */
  g_print("main: calling gtk_main\n");
  gtk_main();

  g_print("main: returned from gtk_main and exiting with success\n");

  return EXIT_SUCCESS;
}
