/**
 * An approach for issuing RPC method calls using the
 * dbus-binding-tool generated wrappers.
 *
 * This maemo code example is licensed under a MIT-style license,
 * that can be found in the file called "License" in the same
 * directory as this file.
 * Copyright (c) 2007 Nokia Corporation. All rights reserved.
 *
 * In order to shorten the code, the code for handling signals has
 * been removed as well as code dealing with value2.
 *
 * This program also demonstrates the inherent complexity when moving
 * from sync to async calls (more problem scenarios to handle). As
 * such, it is NOT suitable to be copy-pasted!
 */

#include <glib.h>
#include <dbus/dbus-glib.h>
#include <stdlib.h>   /* exit */
#include <sys/time.h> /* struct timeval and friends */
#include <time.h>     /* gettimeofday */

/* Pull the common symbolic defines. */
#include "common-defs.h"

/* Pull in the client stubs that were generated with
   dbus-binding-tool */
#include "value-client-stub.h"

/**
 * Print out an error message and optionally quit (if fatal is TRUE)
 */
static void handleError(const char* msg, const char* reason,
                                         gboolean fatal) {
  g_printerr(PROGNAME ": ERROR: %s (%s)\n", msg, reason);
  if (fatal) {
    exit(EXIT_FAILURE);
  }
}

/**
 * Utility to return a pointer to a statically allocated buffer that
 * holds the text representation of seconds since this program was
 * started. Not safe to use in threaded programs!
 */
static const gchar* timestamp(void) {
  /* Holds the "seconds since starting" string. */
  static gchar buffer[200] = {};
  /* Holds the starting timestamp. 0 means that it has not been
     initialized. */
  static guint64 startTimestamp = 0;
  /* Holds the current timestamp. */
  guint64 curTimestamp = 0;
  /* Temp storage for the secs + microseconds time. */
  struct timeval tv;
  /* Temp storage for the difference between start and now. */
  guint64 delta;

  /* Get current time and convert into microseconds flat. */
  gettimeofday(&tv, NULL);
  /* Convert into microseconds. */
  curTimestamp = (guint64)tv.tv_usec + ((guint64)tv.tv_sec)*1000000ULL;

  /* Running for the first time? */
  if (startTimestamp == 0) {
    /* Copy to prev so that we get 0 delta. */
    startTimestamp = curTimestamp;
  }

  /* Calculate the delta (in microseconds). */
  delta = curTimestamp - startTimestamp;

  /* Create the string giving offset from start in seconds. */
  g_snprintf(buffer, sizeof(buffer), "%2u.%.2u",
    (guint)delta / 1000000,
    ((guint)delta % 1000000) / 10000);

  /* Return pointer to the buffer (will always return a pointer to the
     same location. */
  return buffer;
}

/**
 * This function will be called when the async setvalue1 will either
 * complete, timeout or fail (our server however does not signal
 * errors, but the client D-Bus library might). When this example
 * program is left running for a while, you will see all three cases.
 *
 * The prototype must match the one generated by the dbus-binding-tool
 * (org_maemo_Value_setvalue1_reply).
 *
 * Since there is no return value from the RPC, the only useful
 * parameter that we get is the error object, which we'll check.
 * If error is NULL, that means no error. Otherwise the RPC call
 * failed and we should check what the cause was.
 */
static void setValue1Completed(DBusGProxy* proxy, GError *error,
                                                  gpointer userData) {

  g_print(PROGNAME ":%s:setValue1Completed\n", timestamp());
  if (error != NULL) {
    g_printerr(PROGNAME "       ERROR: %s\n", error->message);
    /* We need to release the error object since the stub code does
       not do it automatically. */
    g_error_free(error);
  } else {
    g_print(PROGNAME "       SUCCESS\n");
  }
}

/**
 * This function will be called repeatedly from within the mainloop
 * timer launch code.
 *
 * It will launch asynchronous RPC method to set value1 with ever
 * increasing argument.
 */
static gboolean timerCallback(DBusGProxy* remoteobj) {

  /* Local value that we'll start updating to the remote object. */
  static gint localValue1 = -80;

  /* Start the RPC.
     This is done by calling the stub function that will take the new
     value and the callback function to call on reply getting back.

     The stub returns a DBusGProxyCall object, but we don't need it
     so we'll ignore the return value. The return value could be used
     to cancel a pending request (from client side) with
     dbus_g_proxy_cancel_call. We could also pass a pointer to
     user-data (last parameter), but we don't need one in this example.
     It would normally be used to "carry around" the application state.
     */
  g_print(PROGNAME ":%s:timerCallback launching setvalue1\n",
          timestamp());
  org_maemo_Value_setvalue1_async(remoteobj, localValue1,
                                  setValue1Completed, NULL);
  g_print(PROGNAME ":%s:timerCallback setvalue1 launched\n",
          timestamp());

  /* Step the local value forward. */
  localValue1 += 10;

  /* Repeat timer later. */
  return TRUE;
}

/**
 * The test program itself.
 *
 * 1) Setup GType/GSignal
 * 2) Create GMainLoop object
 * 3) Connect to the Session D-Bus
 * 4) Create a proxy GObject for the remote Value object
 * 5) Start a timer that will launch timerCallback once per second.
 * 6) Run main-loop (forever)
 */
int main(int argc, char** argv) {
  DBusGConnection* bus;
  /* The proxy object. */
  DBusGProxy* remoteValue;
  GMainLoop* mainloop;
  GError* error = NULL;

  g_type_init();

  mainloop = g_main_loop_new(NULL, FALSE);
  /* Failure to create the main loop is fatal (for us). */
  if (mainloop == NULL) {
    handleError("Failed to create the mainloop", "Unknown (OOM?)",
                TRUE);
  }
  g_print(PROGNAME ":main Connecting to Session D-Bus.\n");
  bus = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
  if (error != NULL) {
    handleError("Couldn't connect to the Session bus", error->message,
                TRUE);
  }

  g_print(PROGNAME ":main Creating a GLib proxy object for Value.\n");
  remoteValue =
    dbus_g_proxy_new_for_name(bus,
                              VALUE_SERVICE_NAME, /* name */
                              VALUE_SERVICE_OBJECT_PATH, /* obj path */
                              VALUE_SERVICE_INTERFACE /* interface */);
  if (remoteValue == NULL) {
    handleError("Couldn't create the proxy object",
                "Unknown(dbus_g_proxy_new_for_name)", TRUE);
  }

  g_print(PROGNAME
          ":%s:main Starting main loop (first timer in 1s).\n",
          timestamp());
  g_timeout_add(1000, (GSourceFunc)timerCallback, remoteValue);

  /* Run the program. */
  g_main_loop_run(mainloop);
  /* Since the main loop is not stopped (by this code), we shouldn't
     ever get here. The program might abort() for other reasons. */

  /* If it does, return failure as exit code. */
  return 1;
}
